# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''main module of the cosmology package'''

__version__ = '2022.10.9'

__all__ = [
    'Cosmology',
]

from .background import Cosmology



# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''module for CAMB integration'''

from numbers import Number
import numpy as np

import camb

from .background import Cosmology


def _zarg(*zs):
    za = [z if z is None or isinstance(z, Number) else np.asanyarray(z)
          for z in zs]
    return za if len(za) > 1 else za[0]


class CambCosmology(Cosmology):

    def __init__(self, *args):
        self._p = self._r = None
        for arg in args:
            if isinstance(arg, camb.CAMBparams):
                self._p = arg
            elif isinstance(arg, camb.CAMBdata):
                self._r = arg
            else:
                raise TypeError('argument not CAMB parameters or results')
        if self._p is None and self._r is None:
            raise ValueError('requires CAMB parameters and/or results')
        if self._p is None:
            self._p = self._r.Params
        if self._r is None:
            self._r = camb.get_background(self._p)

    @property
    def h(self) -> float:
        return self._p.h

    @property
    def h0(self) -> float:
        return self._p.H0

    @property
    def omega_m(self) -> float:
        return self._p.omegam

    @property
    def omega_de(self) -> float:
        return self._r.omega_de

    @property
    def omega_k(self) -> float:
        return self._p.omk

    def hf(self, z):
        z = _zarg(z)
        return self._r.hubble_parameter(z)

    def omega_m_z(self, z):
        z = _zarg(z)
        return (self._r.get_Omega('baryon', z) + self._r.get_Omega('cdm', z)
                + self._r.get_Omega('nu', z))

    def omega_de_z(self, z):
        z = _zarg(z)
        return self._r.get_Omega('de', z)

    def omega_k_z(self, z):
        z = _zarg(z)
        return self._r.get_Omega('K', z)

    def dc(self, z, zp=None):
        z, zp = _zarg(z, zp)
        if zp is not None:
            return self._r.comoving_radial_distance(zp) \
                    - self._r.comoving_radial_distance(z)
        return self._r.comoving_radial_distance(z)

    def dc_inv(self, dc):
        return self._r.redshift_at_comoving_radial_distance(dc)

    def da(self, z, zp=None):
        z, zp = _zarg(z, zp)
        if zp is not None:
            return self._r.angular_diameter_distance2(z, zp)
        return self._r.angular_diameter_distance(z)



# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''module for background cosmology'''

from abc import ABCMeta, abstractmethod
import numpy as np


class Cosmology(metaclass=ABCMeta):
    '''Interface for cosmological background calculations.'''

    @classmethod
    def from_astropy(cls, cosmo):
        '''Construct a Cosmology instance from Astropy cosmologies.'''
        from .astropy import AstropyCosmology
        return AstropyCosmology(cosmo)

    @classmethod
    def from_camb(cls, *args):
        '''Construct a Cosmology instance from CAMB.

        Takes CAMB parameters, results, or both, and returns an instance of a
        :class:`Cosmology` adapter class.

        '''

        from .camb import CambCosmology
        return CambCosmology(*args)

    @classmethod
    def _implements(cls, name):
        '''Check whether subclass implements an abstract method.'''
        method = getattr(cls, name, None)
        if method is None:
            return False
        return not getattr(method, '__isabstractmethod__', False)

    @classmethod
    def _default_methods(cls):
        '''List available default methods.'''
        defaults = []
        for c in cls.__mro__:
            for name, value in vars(c).items():
                if name.startswith('_'):
                    parts = name[1:].split('_from_')
                    if len(parts) == 2:
                        method = parts[0]
                        requires = parts[1].split('_and_') if parts[1] else []
                        defaults.append((method, requires, value))
        return defaults

    def __init_subclass__(cls, **kwargs):
        '''Fill out a subclass with available default implementations.'''
        super().__init_subclass__(**kwargs)
        defaults = cls._default_methods()
        for method, requires, default in defaults:
            if not cls._implements(method) \
                    and all(map(cls._implements, requires)):
                setattr(cls, method, default)

    # these are default implementations of methods
    # their names are _{method}_from_{requires}_and_{requires}_and_...
    # their order matters, as they will be checked from top to bottom

    @property
    def _h0_from_hf(self) -> float:
        '''h0 = hf(0)'''
        return self.hf(0)

    @property
    def _h0_from_h(self) -> float:
        '''h0 = 100*h'''
        return self.h*100

    @property
    def _h_from_h0(self) -> float:
        '''h = h0/100'''
        return self.h0/100

    @property
    def _dh_from_h(self) -> float:
        '''dh = 2997.92458 Mpc/h by definition'''
        return 2997.92458/self.h

    @property
    def _grav_from_(self) -> float:
        '''using CODATA 2018 value for Newton's G'''
        return 4.30091727003628e-3

    @property
    def _rho_c_from_grav_and_h(self) -> float:
        return 3e10/(8*np.pi*self.grav)*self.h**2

    def _a_from_(self, z):
        '''a = 1/(1+z)'''
        return 1/np.add(1, z)

    def _ef_from_hf_and_h0(self, z):
        return self.hf(z)/self.h0

    def _ef_from_omega_z(self, z):
        return self.omega_z(z)**0.5

    def _hf_from_ef_and_h0(self, z):
        return self.ef(z)*self.h0

    def _omega_z_from_ef(self, z):
        return self.ef(z)**2

    def _omega_m_z_from_omega_z(self, z):
        return self.omega_m*np.add(1, z)**3/self.omega_z(z)

    def _omega_k_z_from_omega_z(self, z):
        return self.omega_k*np.add(1, z)**2/self.omega_z(z)

    def _omega_de_z_from_omega_m_z_and_omega_k_z(self, z):
        return 1 - self.omega_m_z(z) - self.omega_k_z(z)

    def _rho_c_z_from_rho_c_and_omega_z(self, z):
        return self.rho_c*self.omega_z(z)

    def _rho_m_z_from_rho_c_z_and_omega_m_z(self, z):
        return self.rho_c_z(z)*self.omega_m_z(z)

    def _rho_de_z_from_rho_c_z_and_omega_de_z(self, z):
        return self.rho_c_z(z)*self.omega_de_z(z)

    def _rho_k_z_from_rho_c_z_and_omega_k_z(self, z):
        return self.rho_c_z(z)*self.omega_k_z(z)

    def _dc_from_xc_and_dh(self, z, zp=None):
        return self.xc(z, zp)*self.dh

    def _xc_from_dc_and_dh(self, z, zp=None):
        return self.dc(z, zp)/self.dh

    def _dc_inv_from_xc_inv_and_dh(self, dc):
        return self.xc_inv(dc/self.dh)

    def _xc_inv_from_dc_inv_and_dh(self, xc):
        return self.dc_inv(xc*self.dh)

    def _dm_from_xm_and_dh(self, z, zp=None):
        return self.xm(z, zp)*self.dh

    def _dm_from_da(self, z, zp=None):
        return self.da(z, zp)*np.add(1, z if zp is None else zp)

    def _xm_from_xa(self, z, zp=None):
        return self.xa(z, zp)*np.add(1, z if zp is None else zp)

    def _xm_from_dm_and_dh(self, z, zp=None):
        return self.dm(z, zp)/self.dh

    def _dm_from_dc_and_dh(self, z, zp=None):
        if self.omega_k == 0:
            return self.dc(z, zp)
        if self.omega_k > 0:
            k = self.omega_k**0.5
            return np.sinh(self.dc(z, zp)/self.dh*k)/k*self.dh
        else:
            k = (-self.omega_k)**0.5
            return np.sin(self.dc(z, zp)/self.dh*k)/k*self.dh

    def _xm_from_xc(self, z, zp=None):
        if self.omega_k == 0:
            return self.xc(z, zp)
        if self.omega_k > 0:
            k = self.omega_k**0.5
            return np.sinh(self.xc(z, zp)*k)/k
        else:
            k = (-self.omega_k)**0.5
            return np.sin(self.xc(z, zp)*k)/k

    def _xm_inv_from_dm_inv_and_dh(self, xm):
        return self.dm_inv(xm*self.dh)

    def _dm_inv_from_xm_inv_and_dh(self, dm):
        return self.xm_inv(dm/self.dh)

    def _dm_inv_from_dc_inv_and_dh(self, dm):
        if self.omega_k == 0:
            return self.dc_inv(dm)
        elif self.omega_k > 0:
            k = self.omega_k**0.5
            return self.dc_inv(np.arcsinh(dm/self.dh*k)/k*self.dh)
        else:
            k = (-self.omega_k)**0.5
            return self.dc_inv(np.arcsin(dm/self.dh*k)/k*self.dh)

    def _xm_inv_from_xc_inv(self, xm):
        if self.omega_k == 0:
            return self.xc_inv(xm)
        elif self.omega_k > 0:
            k = self.omega_k**0.5
            return self.xc_inv(np.arcsinh(xm*k)/k)
        else:
            k = (-self.omega_k)**0.5
            return self.xc_inv(np.arcsin(xm*k)/k)

    def _da_from_xa_and_dh(self, z, zp=None):
        return self.xa(z, zp)*self.dh

    def _xa_from_da_and_dh(self, z, zp=None):
        return self.da(z, zp)/self.dh

    def _da_from_dm(self, z, zp=None):
        return self.dm(z, zp)/np.add(1, z if zp is None else zp)

    def _xa_from_xm(self, z, zp=None):
        return self.xm(z, zp)/np.add(1, z if zp is None else zp)

    def _da_inv_from_xa_inv_and_dh(self, da):
        return self.xa_inv(da/self.dh)

    def _da_inv_from_(self, da):
        '''raises NotImplementedError'''
        raise NotImplementedError

    def _xa_inv_from_da_inv_and_dh(self, xa):
        return self.da_inv(xa*self.dh)

    def _dvc_from_xm_and_ef(self, z):
        return self.xm(z)**2/self.ef(z)

    def _vc_from_xc(self, z, zp=None):
        if zp is not None:
            return self.vc(zp) - self.vc(z)
        x = self.xc(z)
        if self.omega_k == 0:
            return x**3/3
        elif self.omega_k > 0:
            k = self.omega_k**0.5
            return (np.sinh(2*k*x) - 2*k*x)/(4*k**3)
        else:
            k = (-self.omega_k)**0.5
            return (2*k*x - np.sin(2*k*x))/(4*k**3)

    # end of default implementations

    @property
    @abstractmethod
    def h(self) -> float:
        '''Dimensionless Hubble parameter :math:`h = H_0/(100 \\, \\rm{km} \\,
        \\rm{s}^{-1} \\, \\rm{Mpc}^{-1})`.'''
        pass

    @property
    @abstractmethod
    def h0(self) -> float:
        '''Hubble function at redshift 0 in km s-1 Mpc-1.'''
        pass

    @property
    @abstractmethod
    def dh(self) -> float:
        '''Hubble distance in Mpc.'''
        pass

    @property
    @abstractmethod
    def omega_m(self) -> float:
        '''Matter density parameter at redshift 0.'''
        pass

    @property
    @abstractmethod
    def omega_de(self) -> float:
        '''Dark energy density parameter at redshift 0.'''
        pass

    @property
    @abstractmethod
    def omega_k(self) -> float:
        '''Curvature density parameter at redshift 0.'''
        pass

    @property
    @abstractmethod
    def rho_c(self) -> float:
        '''Critical density at redshift 0 in Msol Mpc-3.'''
        pass

    @property
    @abstractmethod
    def grav(self) -> float:
        '''Gravitational constant G in pc km2 s-2 Msol-1.'''
        pass

    @abstractmethod
    def a(self, z):
        '''Redshift-dependent scale factor.'''
        pass

    @abstractmethod
    def hf(self, z):
        '''Hubble function :math:`H(z)` in km s-1 Mpc-1.'''
        pass

    @abstractmethod
    def ef(self, z):
        '''Standardised Hubble function :math:`E(z) = H(z)/H_0`.'''
        pass

    @abstractmethod
    def omega_z(self, z):
        '''Redshift-dependent total density parameter.'''
        pass

    @abstractmethod
    def omega_m_z(self, z):
        '''Redshift-dependent matter density parameter.'''
        pass

    @abstractmethod
    def omega_de_z(self, z):
        '''Redshift-dependent dark energy density parameter.'''
        pass

    @abstractmethod
    def omega_k_z(self, z):
        '''Redshift-dependent curvature density parameter.'''
        pass

    @abstractmethod
    def rho_c_z(self, z):
        '''Redshift-dependent critical density in Msol Mpc-3.'''
        pass

    @abstractmethod
    def rho_m_z(self, z):
        '''Redshift-dependent matter density in Msol Mpc-3.'''
        pass

    @abstractmethod
    def rho_de_z(self, z):
        '''Redshift-dependent dark energy density in Msol Mpc-3.'''
        pass

    @abstractmethod
    def rho_k_z(self, z):
        '''Redshift-dependent curvature density in Msol Mpc-3.'''
        pass

    @abstractmethod
    def dc(self, z, zp=None):
        '''Comoving distance :math:`d_c(z)` in Mpc.'''
        pass

    @abstractmethod
    def xc(self, z, zp=None):
        '''Dimensionless comoving distance :math:`x_c'(z) = d_c(z)/d_H`.'''
        pass

    @abstractmethod
    def dc_inv(self, dc):
        '''Inverse function for the comoving distance in Mpc.'''
        pass

    @abstractmethod
    def xc_inv(self, xc):
        '''Inverse function for the dimensionless comoving distance.'''
        pass

    @abstractmethod
    def dm(self, z, zp=None):
        '''Transverse comoving distance :math:`d_M(z)` in Mpc.'''
        pass

    @abstractmethod
    def xm(self, z, zp=None):
        '''Dimensionless transverse comoving distance :math:`x_M(z) =
        d_M(z)/d_H`.'''
        pass

    @abstractmethod
    def dm_inv(self, dm):
        '''Inverse function for the transverse comoving distance in Mpc.'''
        pass

    @abstractmethod
    def xm_inv(self, xm):
        '''Inverse function for the dimensionless transverse comoving
        distance.'''
        pass

    @abstractmethod
    def da(self, z, zp=None):
        '''Angular diameter distance :math:`d_A(z)` in Mpc.'''
        pass

    @abstractmethod
    def xa(self, z, zp=None):
        '''Dimensionless angular diameter distance :math:`x_A(z) =
        d_A(z)/d_H`.'''
        pass

    @abstractmethod
    def da_inv(self, da):
        '''Inverse function for the angular diameter distance in Mpc.'''
        pass

    @abstractmethod
    def xa_inv(self, xa):
        '''Inverse function for the dimensionless angular diameter distance.'''
        pass

    @abstractmethod
    def dvc(self, z):
        r'''Dimensionless differential comoving volume.

        If :math:`V_c` is the comoving volume of a redshift slice with solid
        angle :math:`\Omega`, this function returns

        .. math::

            \mathtt{dvc(z)}
            = \frac{1}{d_H^3} \, \frac{dV_c}{d\Omega \, dz}
            = \frac{x_M^2(z)}{E(z)}
            = \frac{\mathtt{xm(z)^2}}{\mathtt{ef(z)}} \;.

        '''
        pass

    @abstractmethod
    def vc(self, z, zp=None):
        '''Dimensionless comoving volume.'''
        pass



# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''module for large scale structure'''

import numpy as np
from scipy.special import loggamma
from fftl import fftl
from fftl.transforms import sph_hankel

PI = np.pi
SRPI = PI**0.5


def pktoxi(k, pk, q=0.0, kr=1.0, krgood=True):
    '''compute 3d correlation function from power spectrum

    Parameters
    ----------
    k : array_like (N,)
        Wavenumbers at which the power spectrum is given.  Must have
        logarithmic spacing.
    pk : array_like (..., N)
        Power spectrum for given wavenumbers ``k``.  Can be multidimensional.
        Last axis must agree with the wavenumber axis.
    q : float, optional
        Bias parameter for integral transform.
    kr : float, optional
        Shift parameter for logarithmic output grid.
    krgood : bool, optional
        Change given ``kr`` to the nearest value fulfilling the low-ringing
        condition.

    Returns
    -------
    r : array_like (N,)
        Distances at which the correlation function is computed.
    xi : array_like (..., N)
        Correlation values for given distances ``r``.  Leading axes correspond
        to the input power spectrum.

    '''

    r, xi = sph_hankel(0, k, pk, q=q, kr=kr, krgood=krgood)
    xi /= 2*PI**2

    return r, xi


def xitopk(r, xi, q=0.0, kr=1.0, krgood=True):
    '''compute 3d power spectrum from correlation function

    Parameters
    ----------
    r : array_like (N,)
        Distances at which the correlation function is given.  Must have
        logarithmic spacing.
    xi : array_like (..., N)
        Correlation values for given distances ``r``.  Can be multidimensional.
        Last axis must agree with the distance axis.
    q : float, optional
        Bias parameter for integral transform.
    kr : float, optional
        Shift parameter for logarithmic output grid.
    krgood : bool, optional
        Change given ``kr`` to the nearest value fulfilling the low-ringing
        condition.

    Returns
    -------
    k : array_like (N,)
        Wavenumbers at which the power spectrum is compiuted.
    pk : array_like (..., N)
        Power spectrum for given wavenumbers ``k``.  Leading axes correspond to
        the input correlation function.

    '''

    k, pk = sph_hankel(0, r, xi, q=q, kr=kr, krgood=krgood)
    pk *= 4*PI

    return k, pk


def sigma2_r(k, pk, q=0.0, kr=1.0, window='tophat', krgood=True, deriv=False):
    r'''mass variance from matter power spectrum

    Computes the mass variance :math:`\sigma^2(r)` inside a spherical window of
    scale :math:`r` from an input matter power spectrum.  Commonly a tophat
    window is used to produce the variance in spheres of a given radius, but
    other choices are supported.

    The input matter power must be given on a logarithmic grid, and the mass
    variance will returned on a logarithmic grid.  By default, the output grid
    is scaled such that :math:`k_i \, r_{n-i+1} = 1 \forall i = 1, \ldots, n`,
    but can be shifted to other constants using the ``kr`` parameter.

    The function optionally computes the derivative of :math:`\sigma^2(r)` with
    respect to :math:`\ln r` at the same time, if ``deriv`` is true.

    Parameters
    ----------
    k : array_like (N,)
        Wavenumbers at which the power spectrum is given.  Must have
        logarithmic spacing.
    pk : array_like (..., N)
        Power spectrum for given wavenumbers ``k``.  Can be multidimensional.
        Last axis must agree with the wavenumber axis.
    q : float, optional
        Bias parameter for integral transform.
    kr : float, optional
        Shift parameter for logarithmic output grid.
    window : str, optional
        Type of window function for computing the mass variance.  Supported
        values are ``'tophat'``, ``'gaussian'``.
    krgood : bool, optional
        Change given ``kr`` to the nearest value fulfilling the low-ringing
        condition.
    deriv : bool, optional
        Also return the first derivative of the mass variance.

    Returns
    -------
    r : array_like (N,)
        Scales at which the mass variance is evaluated.
    sigma2_r : array_like (..., N)
        Mass variance in spheres of scale ``r``.  Leading axes correspond to
        the input power spectrum.
    dsigma2_dlnr : array_like (..., N), optional
        If ``deriv`` is true, the derivative of ``sigma2_r`` with respect to
        the logarithm of ``r``.

    Notes
    -----
    The mass variance is an integral transform of the matter power spectrum
    :math:`P(k)`,

    .. math::

        \sigma^2_r = \frac{1}{2\pi^2}
                        \int_{0}^{\infty} \! P(k) \, k^2 \, w^2(kr) \, dk \;.

    If :math:`P(k)` is given on a logarithmic grid of :math:`k` values, the
    integral can be computed for a logarithmic grid of :math:`r` values with a
    modification of Hamilton's FFTLog algorithm [1]_,

    .. math::

        U(x) = \int_{0}^{\infty} \! t^x \, w^2(t) \, dt \;.

    The implementation supports the usual tophat window function,

    .. math::

        w(x) = \frac{3}{x^3} \, \bigl\{\sin(x) - x \cos(x)\bigr\} \;,

    and the Gaussian window function,

    .. math::

        w(x) = \exp(-x^2/2) \;.

    References
    ----------
    .. [1] Hamilton A. J. S., 2000, MNRAS, 312, 257.
           doi:10.1046/j.1365-8711.2000.03071.x

    Examples
    --------
    Create a mock power spectrum with a realistic shape.

    >>> k = np.logspace(-4, 2, 40)
    >>> pk = 4e6*k/(1 + k*25)**3.5
    >>>
    >>> import matplotlib.pyplot as plt
    >>> plt.loglog(k, pk)
    >>> plt.xlabel('$k$')
    >>> plt.ylabel('$P(k)$')
    >>> plt.show()

    Compute the mass variance from the power spectrum without setting any
    optional parameters.

    >>> from cosmology import sigma2_r
    >>>
    >>> r, s2 = sigma2_r(k, pk)
    >>>
    >>> plt.loglog(r, s2)
    >>> plt.xlabel('$r$')
    >>> plt.ylabel('$\\sigma^2_r$')
    >>> plt.show()

    The computed mass variance shows a numerical issue on the right, which is
    due to the circular nature of the integral transform it employs.  By
    computing the mass variance with a biased transform (``q = 0.8``), the
    problem disappears.  The exact value of the bias parameter ``q`` depends on
    the shape of the input power spectrum.

    >>> r, s2 = sigma2_r(k, pk, q=0.8)
    >>>
    >>> plt.loglog(r, s2)
    >>> plt.xlabel('$r$')
    >>> plt.ylabel('$\\sigma^2_r$')
    >>> plt.show()

    The integral transform method makes it possible to compute the derivative
    of the mass variance with very little extra computational cost.

    >>> r, s2, ds2_dlnr = sigma2_r(k, pk, q=0.8, deriv=True)
    >>>
    >>> plt.loglog(r, s2)
    >>> plt.loglog(r, -ds2_dlnr)
    >>> plt.xlabel('$r$')
    >>> plt.ylabel('$\\sigma^2_r$ and $-d\\sigma^2_r/d\ln r$')
    >>> plt.show()

    '''

    # window function
    if window == 'tophat':
        if not -1 < q < 3:
            raise ValueError('bias error: tophat window requires -1 < q < 3')

        def u(x):
            dlg = loggamma((1 + x)/2) - loggamma((4 - x)/2)
            return 9*SRPI*np.exp(dlg)/((4 - x)**2 - 1)

    elif window == 'gaussian':
        if not q > -1:
            raise ValueError('bias error: gaussian window requires q > -1')

        def u(x):
            return np.exp(loggamma((x + 1)/2))/2

    else:
        raise ValueError(f'unknown window function: {window}')

    # compute the transform for the given window
    # input function of the transform is pk*k**2 to give a more natural bias
    r, *vals = fftl(u, k, pk*k**2, q=q, kr=kr, krgood=krgood, deriv=deriv)

    # scale all output values by prefactor
    f = 1/(2*PI**2)
    for v in vals:
        v *= f

    # return results
    return (r, *vals)



# author: Nicolas Tessore <n.tessore@ucl.ac.uk>
# license: MIT
'''module for Astropy integration'''

import numpy as np
from astropy.cosmology import z_at_value
from astropy import units as u

from .background import Cosmology


class AstropyCosmology(Cosmology):

    def __init__(self, cosmology):
        self._c = cosmology

    @property
    def h(self) -> float:
        return self._c.h

    @property
    def h0(self) -> float:
        return self._c.H0.value

    @property
    def omega_m(self) -> float:
        return self._c.Om0

    @property
    def omega_de(self) -> float:
        return self._c.Ode0

    @property
    def omega_k(self) -> float:
        return self._c.Ok0

    @property
    def dh(self) -> float:
        return self._c.hubble_distance.to_value('Mpc')

    @property
    def rho_c(self) -> float:
        return self._c.critical_density0.to_value('Msun Mpc-3')

    def a(self, z):
        return self._c.scale_factor(z)

    def hf(self, z):
        return self._c.H(z).to_value('km s-1 Mpc-1')

    def ef(self, z):
        return self._c.efunc(z)

    def omega_z(self, z):
        return self._c.Otot(z)

    def omega_m_z(self, z):
        return self._c.Om(z)

    def omega_de_z(self, z):
        return self._c.Ode(z)

    def omega_k_z(self, z):
        return self._c.Ok(z)

    def rho_c_z(self, z):
        return self._c.critical_density(z).to_value('Msun Mpc-3')

    def dc(self, z, zp=None):
        if zp is not None:
            return self._c.comoving_distance(zp).value \
                    - self._c.comoving_distance(z).value
        return self._c.covoming_distance(z).value

    def dc_inv(self, dc):
        return z_at_value(self._c.comoving_distance, dc*u.Mpc).value

    def dm(self, z, zp=None):
        if zp is not None:
            return self._c.angular_diameter_distance_z1z2(z, zp).value \
                    * np.add(1, zp)
        return self._c.comoving_transverse_distance(z).value

    def dm_inv(self, dm):
        return z_at_value(self._c.comoving_transverse_distance, dm*u.Mpc).value

    def da(self, z, zp=None):
        if zp is not None:
            return self._c.angular_diameter_distance_z1z2(z, zp).value
        return self._c.angular_diameter_distance(z).value

    def da_inv(self, da):
        return z_at_value(self._c.angular_diameter_distance, da*u.Mpc).value
